/*
 * osdthumbview.cpp
 *
 * Copyright (C) 2008 Christophe Thommeret <hftom@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "osdthumbview.h"



OsdThumbView::OsdThumbView( OsdConfig *conf ) : OsdObject( conf )
{
	timeLine.setDuration( 400 );
	connect( &timeLine, SIGNAL(startRepaint()), this, SIGNAL(startRepaint()) );
}

OsdThumbView::~OsdThumbView()
{
	clear();
}

void OsdThumbView::clear()
{
	int i;
	for ( i=0; i<thumbs.count(); ++i )
		delete thumbs.at(i);
	thumbs.clear();
	if ( timeLine.isRunning() ) {
		timeLine.stop();
		emit stopRepaint();
	}
	timeLine.setRange(0);
}

void OsdThumbView::append( OsdThumbItem *it )
{
	thumbs.append( it );
	timeLine.setRange(thumbs.count()-1);
}

void OsdThumbView::updateThumb( const QString &s )
{
	for ( int i=0; i<thumbs.count(); ++i ) {
		if ( thumbs.at(i)->updateThumb( s ) ) {
			emit wantRepaint();
			return;
		}
	}
}

bool OsdThumbView::mouseClicked( QMouseEvent *e )
{
	if ( !inMe( e->x(), e->y() ) || !thumbs.count() || timeLine.isRunning() )
		return false;
	if ( currentThumb.rect.contains( e->x(), e->y() ) ) {
		emit clicked( thumbs.at(timeLine.currentPos())->text() );
		return true;
	}
	return false;
}

void OsdThumbView::wheelEvent( QWheelEvent *e )
{
	if ( !inMe( e->x(), e->y() ) )
		return;

	if ( e->delta()<0 )
		timeLine.stepForward();
	else
		timeLine.stepBackward();
}

void OsdThumbView::paint()
{
	if ( rect.width()<=0 )
		return;

	if ( !thumbs.count() ) {
		return;
	}

	int i, scroll=0;

	if ( timeLine.isRunning() ) {
		scroll+=setCoords();
	}

	scroll+=timeLine.currentPos();

	for ( i=0; i<4; ++i ) {
		if ( (scroll-4+i)>=0 ) {
			glColor4f( 1, 1, 1, leftThumb[ i ].opacity );
			thumbs.at(scroll-4+i)->paint( leftThumb[ i ].rect );
		}
		if ( (scroll+4-i)<thumbs.count() ) {
			glColor4f( 1, 1, 1, rightThumb[ i ].opacity );
			thumbs.at(scroll+4-i)->paint( rightThumb[ i ].rect );
		}
	}

	glColor4f( 1, 1, 1, currentThumb.opacity );
	thumbs.at(scroll)->paint( currentThumb.rect );
	thumbs.at(scroll)->paintText( textRect );

	/*p->fillRect( textRect, config->listItemGrad0Color() );
	p->setPen( Qt::white );
	p->setFont( config->normalFont() );
	p->setOpacity( 1 );
	p->drawText( textRect, Qt::AlignCenter, thumbs.at(scroll)->text() );
	p->setClipping( false );*/
}

void OsdThumbView::resize( int wi, int h )
{
	rect.setHeight( h );
	rect.setWidth( wi );

	QFontMetrics fm( config->normalFont() );
	QRect br = fm.boundingRect( QRect(), Qt::AlignVCenter, "Pepe" );

	float cw = rect.width()-(2*MARGINSIZE);
	float ch = rect.height()-(2*MARGINSIZE);
	float th = br.height();
	while ( cw>(2*(ch-th)) )
		--cw;
	float off = (rect.width()-cw)/2;
	clipRect = QRectF( rect.x()+off, rect.y()+MARGINSIZE, cw, rect.height()-(2*MARGINSIZE) );

	float w = clipRect.width()/3;
	scale = ((clipRect.width()/2)-(2*w/3))/4;
	incx = ((clipRect.width()/2)-((2*w/3)+(scale*3)))/4;
	incy = ((clipRect.y()+clipRect.height()-br.height()-(clipRect.width()/2))-(clipRect.y()+clipRect.height()/2-2*w/3))/4;

	setCoords( true );

	textRect = QRectF( rect.x()+MARGINSIZE, clipRect.y()+clipRect.height()-br.height(), rect.width()-(2*MARGINSIZE), br.height() );
}

int OsdThumbView::setCoords( bool ended )
{
	int i;
	int scrollOffset=0;
	qreal w, w23, ww, xx, yy, wh;
	qreal cx=clipRect.x(), cy=clipRect.y(), cw=clipRect.width(), ch=clipRect.height();
	qreal val;
	if ( ended ) {
		w = cw/3;
		w23 = 2*w/3;
		for ( i=0; i<4; ++i ) {
			leftThumb[ i ].rect = QRectF(cx+(i*incx), (cy+ch/2-w23)+(i*incy), w23+(scale*i), (w23+(scale*i))*3/4);
			leftThumb[ i ].opacity = (i+1)*0.2;
			rightThumb[ i ].rect = QRectF(cx+cw-w23-(i*incx)-(scale*i), (cy+ch/2-w23)+(i*incy), w23+(scale*i), (w23+(scale*i))*3/4);
			rightThumb[ i ].opacity = (i+1)*0.2;
		}
		w = cw/2;
		currentThumb.rect = QRectF( cx+w/2, cy+ch-textRect.height()-w, w, w*3/4 );
		currentThumb.opacity = 1.0;
		currentRightInc = rightThumb[3].rect.x()-currentThumb.rect.x();
		currentLeftInc = currentThumb.rect.x()-leftThumb[3].rect.x();
		currentScale = currentThumb.rect.width()-leftThumb[ 3 ].rect.width();
		return 0;
	}

	val = timeLine.currentValue();
	if ( val==-1.0 ) {
		finished();
		return 0;
	}

	if ( val<0.5 ) {
		val*=2;
		w = cw/2;
		if ( timeLine.getDirection()==1 ) {
			scrollOffset=0;
			ww = w-(val*currentScale);
			xx = w/2-val*currentLeftInc;
			yy = -incy*val;
		}
		else {
			scrollOffset=-1;
			ww = w-(val*currentScale);
			xx = w/2-val*currentLeftInc;
			yy = -incy*val;
		}
		currentThumb.rect = QRectF( cx+xx, cy+ch-textRect.height()-w+yy, ww, ww*3/4 );
		currentThumb.opacity = 1.0-(0.2*val);
		w = cw/3;
		w23 = 2*w/3;
		for ( i=0; i<4; ++i ) {
			wh = w23+(scale*(i-val));
			leftThumb[ i ].rect = QRectF(cx+(i*incx)-(val*incx), (cy+ch/2-w23)+(i*incy)-(val*incy), wh, wh*3/4);
			leftThumb[ i ].opacity = (i+1-val)*0.2;

			rightThumb[ i ].rect = QRectF(cx+cw-w23-(i*incx)-(scale*i), (cy+ch/2-w23)+(i*incy), w23+(scale*i), (w23+(scale*i))*3/4);
			rightThumb[ i ].opacity = (i+1)*0.2;
		}

	}
	else {
		val-=0.5;
		val*=2;
		w = cw/2;
		if ( timeLine.getDirection()==1 ) {
			scrollOffset=1;
			ww = w-currentScale+(val*currentScale);
			xx = w/2+currentRightInc-val*currentRightInc;
			yy = (incy*val)-incy;
		}
		else {
			scrollOffset=0;
			ww = w-currentScale+(val*currentScale);
			xx = w/2+currentRightInc-val*currentRightInc;
			yy = (incy*val)-incy;
		}
		currentThumb.rect = QRectF( cx+xx, cy+ch-textRect.height()-w+yy, ww, ww*3/4 );
		currentThumb.opacity = 0.8+(0.2*val);
		w = cw/3;
		w23 = 2*w/3;
		for ( i=0; i<4; ++i ) {
			leftThumb[ i ].rect = QRectF(cx+(i*incx), (cy+ch/2-w23)+(i*incy), w23+(scale*i), (w23+(scale*i))*3/4);
			leftThumb[ i ].opacity = (i+1)*0.2;

			wh = w23+(scale*(i-1+val));
			rightThumb[ i ].rect = QRectF(cx+cw-wh-(i*incx)+incx-(val*incx), (cy+ch/2-w23)+(i*incy)-incy+(val*incy), wh, wh*3/4);
			rightThumb[ i ].opacity = (i+val)*0.2;
		}
	}
	return scrollOffset;
}

void OsdThumbView::finished()
{
	setCoords( true );
	timeLine.stop();
	stopRepaint();
}
